#!/usr/bin/python3
#
# Mastostream, a command-line Mastodon client, just for keeping up.
# Shows your timeline, warns you about notifications, remembers where you got to.
# Allows you to view threads, bookmark, favourite, and vote, but that's it.
# (niceties: working CWs, external media viewing, media captions, ASCII art polls)
# (problems: untested voting, filters are broken unless the server handles them)
#
# Copyright 2021 Fudge the Sphinx
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.

import os
import re
import requests
import subprocess
import sys
import tempfile
import time

# config
scheme = "https"
instance = "some-instance.com" # TODO: fill this in!
client_name = "mastostream"
client_website = None
version = "0.1"
# letters used to name poll options when displaying and voting
poll_binds = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z"]
# key binds that trigger media file download and view
media_binds = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
# programs used to view media files, arguments should given as separate strings in each list
# the path to the file is appended to the list and the result is given to subprocess.run()
media_commands = {
	"image": ["gpicview"], # TODO: fill this in!
	"gifv": ["mplayer"], # TODO: fill this in!
	"video": ["mplayer"], # TODO: fill this in!
	"audio": ["mplayer"], # TODO: fill this in!
}
# file downloads abort if the file is longer than this to avoid wasting memory and bandwidth
max_download_bytes = 10000000

# constants
timeout = 30
scopes = "read:favourites write:favourites read:bookmarks write:bookmarks read:notifications write:notifications read:statuses write:statuses" # need write:statuses to vote in polls O.o

# state
access_token = None

# keypress setup
wait_for_keypress = None
if sys.platform == "win32":
	# Untested!
	import msvcrt
	wait_for_keypress = msvcrt.getch
else:
	# Only tested on Linux!
	import termios
	def fn():
		fd = sys.stdin.fileno()
		termios.tcflush(fd, termios.TCIFLUSH)
		attributes = termios.tcgetattr(fd)
		changes = termios.tcgetattr(fd)
		if changes[3]&termios.ICANON: changes[3] ^= termios.ICANON
		changes[6][termios.VMIN] = 1
		changes[6][termios.VTIME] = 0
		termios.tcsetattr(fd, termios.TCSANOW, changes)
		try:
			return sys.stdin.read(1)
		finally:
			termios.tcsetattr(sys.stdin, termios.TCSANOW, attributes)
	wait_for_keypress = fn

# wait for a keypress and try to ignore accidental pastes of text into the terminal
last_key_time = None
def wait_for_keypress_no_paste():
	global last_key_time
	minimum_seconds = 0.1
	while True:
		key = wait_for_keypress()
		now = time.monotonic()
		ok = last_key_time == None or now-last_key_time >= minimum_seconds
		last_key_time = now
		if ok:
			return key
		print(prefix_error+"input too fast, paste protection triggered")

# colour setup
style_default = ""
style_faint = ""
style_ok = ""
style_warning = ""
style_error = ""
if sys.platform != "win32" or "ANSICON" in os.environ:
	style_default = "\033[0m"
	style_faint = "\033[90m"
	style_ok = "\033[32m"
	style_warning = "\033[33m"
	style_error = "\033[31m"
prefix_ok = style_ok+"[ok] "+style_default
prefix_warning = style_warning+"[??] "+style_default
prefix_error = style_error+"[!!] "+style_default
prefix_prompt = style_ok+"[>>] "+style_default

def get_config_directory():
	home = None
	if sys.platform == "win32" and "APPDATA" in os.environ:
		home = os.environ["APPDATA"]
	elif "XDG_CONFIG_HOME" in os.environ:
		home = os.path.expanduser(os.environ["XDG_CONFIG_HOME"])
	else:
		home = os.path.join(os.path.expanduser("~"), ".config")
	return os.path.join(home, client_name)

def save_config_item(name, value):
	directory = get_config_directory()
	os.makedirs(directory, exist_ok=True)
	path = os.path.join(directory, name)
	fd = os.open(path, os.O_CREAT | os.O_WRONLY, 0o600) # token file contains secrets
	with os.fdopen(fd, "w") as file: file.write(value)

def load_config_item(name):
	directory = get_config_directory()
	path = os.path.join(directory, name)
	try:
		with open(path, "r") as file: return file.read()
	except OSError:
		return None

def get_headers():
	headers = {"User-Agent": "{}/{}".format(client_name, version)}
	if access_token: headers["Authorization"] = "Bearer {}".format(access_token)
	return headers

def do_get(path, params):
	assert(path[0] == '/')
	url = "{}://{}{}".format(scheme, instance, path)
	response = requests.request("GET", url, params=params, timeout=timeout, allow_redirects=False, headers=get_headers())
	response.raise_for_status()
	return response.json()

def do_post(path, params):
	assert(path[0] == '/')
	url = "{}://{}{}".format(scheme, instance, path)
	response = requests.request("POST", url, data=params, timeout=timeout, allow_redirects=False, headers=get_headers())
	response.raise_for_status()
	return response.json()

def do_download(url):
	bar_width = 10
	with requests.request("GET", url, timeout=timeout, allow_redirects=True, headers=get_headers(), stream=True) as response:
		response.raise_for_status()
		length = response.headers.get("content-length")
		if length:
			length = int(length)
			if length < 0: length = 1
			if length > max_download_bytes:
				print(prefix_error+"file too large, cancelling download")
				return None
		(fd, path) = tempfile.mkstemp()
		failed = False
		with os.fdopen(fd, mode="wb") as file:
			iter = response.iter_content(chunk_size=None)
			try:
				while True:
					bytes = response.raw.tell()
					if bytes > max_download_bytes:
						print(prefix_error+"file too large, cancelling download")
						failed = True
						break
					if length:
						assert(bytes <= length)
						bars = round(bytes/length*bar_width)
						print("{}[{}{}] downloading...\r".format(style_default, "|"*bars, " "*(bar_width-bars)), end="")
					else:
						print("{}[{}k] downloading...\r".format(style_default, round(bytes/1000)), end="")
					sys.stdout.flush()
					try:
						data = next(iter)
						file.write(data)
					except StopIteration:
						break
			except KeyboardInterrupt:
				print(prefix_error+"cancelled                  ")
				failed = True
		if failed:
			os.remove(path)
			return None
		print(prefix_ok+"downloaded                 ")
		return path

def greet():
	# check instance is alive
	info = do_get("/api/v1/instance", None)
	print("{}welcome to {}".format(prefix_ok, info["title"]))

def authorize():
	global access_token
	print(prefix_warning+"no saved access token, press enter to login")
	input(prefix_prompt)

	# register client
	out_of_band = "urn:ietf:wg:oauth:2.0:oob"
	data = {
		"client_name": client_name,
		"redirect_uris": out_of_band,
		"scopes": scopes,
	}
	if client_website: data["website"] = client_website
	info = do_post("/api/v1/apps", data)
	id = info["client_id"]
	secret = info["client_secret"]
	print(prefix_ok+"register client")

	# ask user to log in
	params = [
		"client_id={}".format(id),
		"scope={}".format(scopes.replace(" ", "+")),
		"redirect_uri={}".format(out_of_band),
		"response_type=code",
	]
	query = "&".join(params).replace(":", "%3A")
	print(prefix_prompt+"please login and paste resulting code")
	print("{}{}://{}/oauth/authorize?{}".format(prefix_prompt, scheme, instance, query))
	code = ""
	while code == "": code = input(prefix_prompt)

	# get access token
	info = do_post("/oauth/token", {
		"client_id": id,
		"client_secret": secret,
		"redirect_uri": out_of_band,
		"scope": scopes,
		"code": code,
		"grant_type": "authorization_code",
	})
	access_token = info["access_token"]
	assert(info["token_type"] == "Bearer")
	print(prefix_ok+"log in")

	# store token
	save_config_item("token", access_token)
	print(prefix_ok+"access token saved")

re_break = re.compile(r"(?:</?p[^>]*>)|(?:<br */?>)")
re_tag = re.compile(r"</?[^>]+>")
def html_to_plain(html):
	broken = "\n".join([p for p in re_break.split(html) if p])
	return re_tag.sub("", broken).replace("&gt;", ">").replace("&lt;", "<").replace("&amp;", "&").replace("&apos;", "'").replace("&quot;", "\"")

def print_poll(poll):
	assert(len(poll["options"]) <= len(poll_binds))
	bar_width = 10

	if poll["voted"] or poll["expired"]:
		max = 0
		for option in poll["options"]:
			if option["votes_count"] > max: max = option["votes_count"]
		for (index, option) in enumerate(poll["options"]):
			bars = 0 if max == 0 else round(option["votes_count"]/max*bar_width)
			print("{}{} {}[{}{}] {}".format(style_faint, poll_binds[index], style_default, "|"*bars, " "*(bar_width-bars), option["title"]))
	else:
		for (index, option) in enumerate(poll["options"]):
			print("{}{} {}{}".format(style_faint, poll_binds[index], style_default, option["title"]))
	people = poll["voters_count"] or poll["votes_count"]
	print("{}{} people".format(style_faint, people))

def print_media(media_list, is_sensitive):
	assert(len(media_list) <= len(media_binds))
	style = style_error if is_sensitive else style_default
	for (index, media) in enumerate(media_list):
		label = media["type"]
		if is_sensitive: label = "sensitive "+label
		print("{}{} {}[{}]".format(style_faint, media_binds[index], style, label))

def print_status(status, show_sensitive):
	assert(status)

	# display info for actual post
	is_reblog = status["reblog"] != None
	forwarder = status["account"]
	if is_reblog: status = status["reblog"]
	author = status["account"]

	# choose visibility icon
	if status["visibility"] == "public":
		visibility_icon = style_ok+"☼"
	elif status["visibility"] == "unlisted":
		visibility_icon = style_ok+"☀"
	elif status["visibility"] == "private":
		visibility_icon = style_warning+"⨯"
	elif status["visibility"] == "direct":
		visibility_icon = style_error+"✉"
	else:
		assert(False)

	# choose reply icon
	is_reply = status["in_reply_to_id"] != None
	has_reply = int(status["replies_count"]) > 0
	if has_reply and is_reply:
		reply_icon = "⇕"
	elif has_reply:
		reply_icon = "⇓"
	elif is_reply:
		reply_icon = "⇑"
	else:
		reply_icon = "-"
	reply_icon = style_faint+reply_icon

	# format author handle
	author_info = "{}{} {}{}".format(style_default, author["display_name"], style_faint, author["acct"])

	# format post time
	posted_at = time.mktime(time.strptime(status["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ")) # force local time, I hope
	elapsed_seconds = time.time() - posted_at
	if elapsed_seconds < 0:
		time_info = "{}[the future]".format(style_ok)
	elif elapsed_seconds < 60:
		time_info = "{}[{}s ago]".format(style_ok, int(elapsed_seconds))
	elif elapsed_seconds < 60*60:
		time_info = "{}[{}m ago]".format(style_ok, int(elapsed_seconds/60))
	elif elapsed_seconds < 60*60*24:
		time_info = "{}[{}h ago]".format(style_warning, int(elapsed_seconds/60/60))
	else:
		time_info = "{}[{}]".format(style_error, time.strftime("%Y-%m-%d", time.localtime(posted_at)))

	# format forwarder info
	if is_reblog:
		forwarder_info = "\n{}↡ {}{} {}{}".format(style_ok, style_default, forwarder["display_name"], style_faint, forwarder["acct"])
	else:
		forwarder_info = ""

	# display post
	print("{}{} {} {}{}".format(visibility_icon, reply_icon, author_info, time_info, forwarder_info))
	if status["spoiler_text"] and not show_sensitive:
		print("{}[{}]".format(style_error, status["spoiler_text"]))
	else:
		print(style_default+html_to_plain(status["content"]))
		if status["poll"]: print_poll(status["poll"])
		print_media(status["media_attachments"], status["sensitive"])

def view_media(url, description, type):
	assert(access_token)
	command = media_commands.get(type)
	if not command:
		print(prefix_error+"unknown media type")
		return
	if description:
		print(style_default+description)
	else:
		print(style_faint+"[uncaptioned]")
	path = do_download(url)
	if not path: return
	subprocess.run(command+[path], check=True)
	os.remove(path)

def vote_in_poll(status_id, poll):
	assert(access_token)
	is_multiple = poll["multiple"]
	count = len(poll["options"])
	plural = "s" if is_multiple else ""
	message = "{}please enter vote letter{}".format(prefix_prompt, plural)
	while True:
		try:
			print(message)
			votes = input(prefix_prompt).upper()
		except KeyboardInterrupt:
			print("\n{}cancelled".format(prefix_error))
			return
		indices = []
		failed = False
		for vote in votes:
			try:
				index = poll_binds.index(vote)
			except ValueError:
				print("{}choice {} invalid".format(prefix_error, vote))
				failed = True
				break
			if index >= count:
				print("{}choice {} not available".format(prefix_error, vote))
				failed = True
				break
			indices.append(index)
		if failed: continue
		if len(indices) != 1 and not is_multiple:
			print("{}need exactly one vote (found {})".format(prefix_error, len(indices)))
			continue
		break
	if len(indices) == 0:
		print(style_faint+"Voting for: nothing")
	else:
		print(style_faint+"Voting for:")
		for index in indices: print("{} {}".format(poll_binds[index], poll["options"][index]["title"]))
	do_post("/api/v1/polls/{}/votes".format(status_id), {"choices": indices})
	print(prefix_ok+"voted")

def view_thread(start_status):
	assert(access_token)
	info = do_get("/api/v1/statuses/{}/context".format(start_status["id"]), {})
	statuses = info["ancestors"]+[start_status]+info["descendants"]
	for status in statuses:
		if status_ui(status, False): return
	print(prefix_ok+"end of thread, press enter to return to normal view")
	input(prefix_prompt)

def status_ui(status, allow_threadview):
	print_status(status, False)
	# action logic
	data = status
	if data["reblog"]: data = data["reblog"]
	show_sensitive = False
	repeat = True
	quit = False
	def action_next(key):
		nonlocal repeat
		repeat = False
	def action_quit(key):
		nonlocal quit
		quit = True
	def action_show(key):
		nonlocal show_sensitive
		show_sensitive = True
		print_status(status, True)
	def action_favourite(key):
		do_post("/api/v1/statuses/{}/favourite".format(data["id"]), None)
		print(prefix_ok+"favourited")
	def action_bookmark(key):
		do_post("/api/v1/statuses/{}/bookmark".format(data["id"]), None)
		print(prefix_ok+"bookmarked")
	def action_info(key):
		print("Post details")
		print("ID: {}".format(data["id"]))
		print("URL: {}".format(data["url"]))
		print("favourites: {}".format(data["favourites_count"]))
		print("reblogs: {}".format(data["reblogs_count"]))
	def action_thread(key):
		view_thread(data)
		print_status(status, show_sensitive)
	def action_vote(key):
		print(prefix_error+"TODO untested!")
		vote_in_poll(data["id"], data["poll"])
	def action_view(key):
		index = media_binds.index(key)
		if index >= len(data["media_attachments"]):
			print(prefix_error+"no such media")
			return
		media = data["media_attachments"][index]
		view_media(media["url"], media["description"], media["type"])
	# choose action
	while repeat and not quit:
		actions = [
			# key, tooltip,  callback
			(" ", "[␣]next", action_next),
			("q", "[q]uit", action_quit),
		]
		if data["spoiler_text"] and not show_sensitive:
			actions.append(("\n", "[⏎]show", action_show))
		else:
			actions.append(("i", "[i]nfo", action_info))
			is_thread = data["in_reply_to_id"] or int(data["replies_count"]) > 0
			if is_thread and allow_threadview: actions.append(("t", "[t]hread", action_thread))
			if not data["favourited"]: actions.append(("f", "[f]avourite", action_favourite))
			if not data["bookmarked"]: actions.append(("b", "[b]ookmark", action_bookmark))
			if data["poll"]: actions.append(("v", "[v]ote", action_vote))
			if len(data["media_attachments"]) > 0: actions.append((None, "[0]-[9]view", action_view))
		instructions = " ".join(map(lambda action: action[1], actions))
		print("\n  {}{}\r{}> ".format(style_faint, instructions, style_default), end="")
		sys.stdout.flush()
		key = wait_for_keypress_no_paste()
		print(" "*len(instructions)+"\r", end="") # overwrite input echo
		matched = False
		for action in actions:
			if action[0] == key or (not action[0] and key in media_binds):
				matched = True
				action[2](key)
		if not matched: print(prefix_error+"unknown command")
	return quit

# Gets statuses from the timeline with a queue so that they can be fetched in batches
status_queue = []
def get_next_status(min_id=None):
	global status_queue
	# Refill queue when empty
	if len(status_queue) == 0:
		params = {}
		if min_id:
			params["min_id"] = min_id
		status_queue = do_get("/api/v1/timelines/home", params)
	if len(status_queue) == 0:
		return None
	else:
		return status_queue.pop()

def next_status():
	assert(access_token)
	last_read = load_config_item("last_read")
	if last_read == None:
		print(prefix_warning+"no last read post, starting from current")
	status = get_next_status(last_read)
	if status == None:
		print(prefix_ok+"no new posts")
		return True
	result = status_ui(status, True)
	# only save once we know all transactions completed successfully
	last_read = status["id"]
	save_config_item("last_read", last_read)
	return result

def check_notifications():
	assert(access_token)
	last_read = load_config_item("last_notification")
	limit = 20
	params = {"limit": limit}
	if last_read:
		params["min_id"] = last_read
	else:
		print(prefix_warning+"no last read notification, starting from oldest")
	cap = 999
	count = 0
	while count < cap:
		info = do_get("/api/v1/notifications", params)
		if len(info) == 0: break
		last_read = info[0]["id"]
		count += len(info)
		if len(info) < limit:
			break
		params["min_id"] = last_read
	if count == 0:
		return
	elif count <= cap:
		print("{}{} new notifications, press enter to continue".format(prefix_warning, count))
	else:
		print("{}>{} new notifications, press enter to continue".format(prefix_error, cap))
	input(prefix_prompt)
	assert(last_read)
	save_config_item("last_notification", last_read)

# for debugging
# def show_status(id):
# 	assert(access_token)
# 	info = do_get("/api/v1/statuses/{}".format(id), {})
# 	status_ui(info, True)

if sys.stdout.isatty():
	greet()
	access_token = load_config_item("token")
	if not access_token: authorize()
	check_notifications()
	while True:
		if next_status(): break
else:
	print(prefix_error+"output is not a shell, this program only supports interactive use")
